@@ -273,6 +273,7 @@ void Decoration::init()
     connect(s.get(), &KDecoration3::DecorationSettings::reconfigured, SettingsProvider::self(), &SettingsProvider::reconfigure, Qt::UniqueConnection);
     connect(s.get(), &KDecoration3::DecorationSettings::reconfigured, this, &Decoration::updateButtonsGeometryDelayed);
 
+    connect(window(), &KDecoration3::DecoratedWindow::activeChanged, this, &Decoration::recalculateBorders);
     connect(window(), &KDecoration3::DecoratedWindow::adjacentScreenEdgesChanged, this, &Decoration::recalculateBorders);
     connect(window(), &KDecoration3::DecoratedWindow::maximizedHorizontallyChanged, this, &Decoration::recalculateBorders);
     connect(window(), &KDecoration3::DecoratedWindow::maximizedVerticallyChanged, this, &Decoration::recalculateBorders);
@@ -404,11 +405,32 @@ void Decoration::recalculateBorders()
     setBorders(bordersFor(window()->nextScale()));
 
     // extended sizes
-    const qreal extSize = window()->snapToPixelGrid(settings()->largeSpacing());
+    const qreal extSize = KDecoration3::snapToPixelGrid(settings()->largeSpacing(), window()->nextScale());
     qreal extSides = 0;
     qreal extBottom = 0;
 
     setResizeOnlyBorders(QMarginsF(extSides, 0, extSides, extBottom));
+
+    qreal bottomLeftRadius = 0;
+    qreal bottomRightRadius = 0;
+    setBorderRadius(KDecoration3::BorderRadius(0, 0, bottomRightRadius, bottomLeftRadius));
+
+    if (isMaximized() || !outlinesEnabled()) {
+        setBorderOutline(KDecoration3::BorderOutline());
+    } else {
+        const auto color = KColorUtils::mix(window()->color(window()->isActive() ? ColorGroup::Active : ColorGroup::Inactive, ColorRole::Frame),
+                                            window()->palette().text().color(),
+                                            lookupOutlineIntensity(m_internalSettings->outlineIntensity()));
+        const qreal thickness = std::max(KDecoration3::pixelSize(window()->scale()), KDecoration3::snapToPixelGrid(1, window()->scale()));
+
+        qreal bottomLeftRadius = 0;
+        qreal bottomRightRadius = 0;
+        bottomLeftRadius = m_scaledCornerRadius;
+        bottomRightRadius = m_scaledCornerRadius;
+
+        const auto radius = KDecoration3::BorderRadius(m_scaledCornerRadius, m_scaledCornerRadius, bottomRightRadius, bottomLeftRadius);
+        setBorderOutline(KDecoration3::BorderOutline(thickness, color, radius));
+    }
 }
 
 //________________________________________________________________
@@ -511,53 +533,12 @@ void Decoration::paint(QPainter *painter, const QRectF &repaintRegion)
         painter->drawRect(rect().adjusted(0, 0, -1, -1));
         painter->restore();
     }
-    if (outlinesEnabled() && !isMaximized()) {
-        auto outlineColor = KColorUtils::mix(window()->color(window()->isActive() ? ColorGroup::Active : ColorGroup::Inactive, ColorRole::Frame),
-                                             window()->palette().text().color(),
-                                             lookupOutlineIntensity(m_internalSettings->outlineIntensity()));
-
-        const qreal lineWidth = std::max(window()->pixelSize(), window()->snapToPixelGrid(1));
-        QRectF outlineRect = rect();
-        // this is necessary to paint exactly in the center of the intended line
-        // otherwise anti aliasing kicks in and ruins the day
-        outlineRect.adjust(lineWidth / 2, lineWidth / 2, -lineWidth / 2, -lineWidth / 2);
-
-        qreal cornerSize = m_scaledCornerRadius * 2.0;
-        QRectF cornerRect(outlineRect.x(), outlineRect.y(), cornerSize, cornerSize);
-
-        QPainterPath outlinePath;
-        outlinePath.arcMoveTo(cornerRect, 180);
-        outlinePath.arcTo(cornerRect, 180, -90);
-        cornerRect.moveTopRight(outlineRect.topRight());
-        outlinePath.arcTo(cornerRect, 90, -90);
-        // Check if border size is "no borders" or "no side-borders"
-        cornerRect.moveBottomRight(outlineRect.bottomRight());
-        outlinePath.arcTo(cornerRect, 0, -90);
-        cornerRect.moveBottomLeft(outlineRect.bottomLeft());
-        outlinePath.arcTo(cornerRect, 270, -90);
-        outlinePath.closeSubpath();
-
-        painter->save();
-        painter->setPen(QPen(outlineColor, lineWidth));
-        painter->setBrush(Qt::NoBrush);
-        painter->setRenderHint(QPainter::Antialiasing);
-        painter->drawPath(outlinePath.simplified());
-        painter->restore();
-    }
 }
 
 //________________________________________________________________
 void Decoration::paintTitleBar(QPainter *painter, const QRectF &repaintRegion)
 {
-    const qreal halfAPixel = 0.5 * window()->pixelSize();
     QRectF rect(QPointF(0, 0), QSizeF(size().width(), borderTop()));
-    const bool noOutlines = isMaximized() || !settings()->isAlphaChannelSupported();
-    const bool bottomOutline = window()->isShaded();
-    if (!noOutlines && outlinesEnabled()) {
-        // this is needed to match the corner radius with the outline
-        rect.adjust(halfAPixel, halfAPixel, -halfAPixel, bottomOutline ? -halfAPixel : 0);
-    }
-
     QBrush frontBrush;
     QBrush backBrush(this->titleBarColor());
 
@@ -582,13 +563,13 @@ void Decoration::paintTitleBar(QPainter *painter, const QRectF &repaintRegion)
         painter->setBrush(titleBarColor());
     }
 
-    if (noOutlines) {
+    if (isMaximized() || !settings()->isAlphaChannelSupported()) {
         painter->setBrush(backBrush);
         painter->drawRect(rect);
 
         painter->setBrush(frontBrush);
         painter->drawRect(rect);
-    } else if (bottomOutline) {
+    } else if (window()->isShaded()) {
         painter->setBrush(backBrush);
         painter->drawRoundedRect(rect, m_scaledCornerRadius, m_scaledCornerRadius);
 
@@ -666,15 +647,17 @@ QPair<QRectF, Qt::Alignment> Decoration::captionRect() const
         return qMakePair(QRectF(), Qt::AlignCenter);
     } else {
         const qreal leftOffset =
-            window()->snapToPixelGrid(m_leftButtons->buttons().isEmpty() ? Metrics::TitleBar_SideMargin * settings()->smallSpacing()
-                                                                         : m_leftButtons->geometry().x() + m_leftButtons->geometry().width()
-                                              + Metrics::TitleBar_SideMargin * settings()->smallSpacing());
+            KDecoration3::snapToPixelGrid(m_leftButtons->buttons().isEmpty() ? Metrics::TitleBar_SideMargin * settings()->smallSpacing()
+                                                                             : m_leftButtons->geometry().x() + m_leftButtons->geometry().width()
+                                                  + Metrics::TitleBar_SideMargin * settings()->smallSpacing(),
+                                          window()->scale());
 
-        const qreal rightOffset = window()->snapToPixelGrid(m_rightButtons->buttons().isEmpty() ? Metrics::TitleBar_SideMargin * settings()->smallSpacing()
-                                                                                                : size().width() - m_rightButtons->geometry().x()
-                                                                    + Metrics::TitleBar_SideMargin * settings()->smallSpacing());
+        const qreal rightOffset = KDecoration3::snapToPixelGrid(m_rightButtons->buttons().isEmpty() ? Metrics::TitleBar_SideMargin * settings()->smallSpacing()
+                                                                                                    : size().width() - m_rightButtons->geometry().x()
+                                                                        + Metrics::TitleBar_SideMargin * settings()->smallSpacing(),
+                                                                window()->scale());
 
-        const qreal yOffset = window()->snapToPixelGrid(settings()->smallSpacing() * Metrics::TitleBar_TopMargin);
+        const qreal yOffset = KDecoration3::snapToPixelGrid(settings()->smallSpacing() * Metrics::TitleBar_TopMargin, window()->scale());
         const QRectF maxRect(leftOffset, yOffset, size().width() - leftOffset - rightOffset, captionHeight());
 
         switch (m_internalSettings->titleAlignment()) {
@@ -732,7 +715,6 @@ void Decoration::updateShadow()
 
     auto &shadow = (window()->isActive()) ? g_sShadow : g_sShadowInactive;
     if (!shadow) {
-        // Update both active and inactive shadows so outline stays consistent between the two
         g_sShadow = createShadowObject(1.0);
         g_sShadowInactive = createShadowObject(0.5);
     }
@@ -804,7 +786,7 @@ void Decoration::setScaledCornerRadius()
     // On X11, the smallSpacing value is used for scaling.
     // On Wayland, this value has constant factor of 2.
     // Removing it will break radius scaling on X11.
-    m_scaledCornerRadius = window()->snapToPixelGrid(Metrics::Frame_FrameRadius * settings()->smallSpacing());
+    m_scaledCornerRadius = KDecoration3::snapToPixelGrid(Metrics::Frame_FrameRadius * settings()->smallSpacing(), window()->nextScale());
 }
 
 void Decoration::updateScale()
